@@ -622,7 +622,7 @@ public final class HttpServerExchange extends AbstractAttachable {
      */
     public String getHostName() {
         String host = requestHeaders.getFirst(Headers.HOST);
-        if (host == null) {
+        if (host == null || "".equals(host.trim())) {
             host = getDestinationAddress().getHostString();
         } else {
             if (host.startsWith("[")) {
@@ -1139,7 +1139,7 @@ public final class HttpServerExchange extends AbstractAttachable {
      * @param cookie The cookie
      */
     public HttpServerExchange setResponseCookie(final Cookie cookie) {
-        if(getConnection().getUndertowOptions().get(UndertowOptions.ENABLE_RFC6265_COOKIE_VALIDATION, UndertowOptions.DEFAULT_ENABLE_RFC6265_COOKIE_VALIDATION)) {
+        if (getConnection().getUndertowOptions().get(UndertowOptions.ENABLE_RFC6265_COOKIE_VALIDATION, UndertowOptions.DEFAULT_ENABLE_RFC6265_COOKIE_VALIDATION)) {
             if (cookie.getValue() != null && !cookie.getValue().isEmpty()) {
                 Rfc6265CookieSupport.validateCookieValue(cookie.getValue());
             }
@@ -1405,6 +1405,11 @@ public final class HttpServerExchange extends AbstractAttachable {
         if (allAreSet(oldVal, FLAG_RESPONSE_SENT)) {
             throw UndertowMessages.MESSAGES.responseAlreadyStarted();
         }
+        if(statusCode >= 500) {
+            if(UndertowLogger.ERROR_RESPONSE.isDebugEnabled()) {
+                UndertowLogger.ERROR_RESPONSE.debugf(new RuntimeException(), "Setting error code %s for exchange %s", statusCode, this);
+            }
+        }
         this.state = oldVal & ~MASK_RESPONSE_CODE | statusCode & MASK_RESPONSE_CODE;
         return this;
     }
@@ -1561,8 +1566,10 @@ public final class HttpServerExchange extends AbstractAttachable {
     }
 
     /**
-     *
-     * @return The request start time, or -1 if this was not recorded
+     * @return The request start time using the JVM's high-resolution time source,
+     * in nanoseconds, or -1 if this was not recorded
+     * @see UndertowOptions#RECORD_REQUEST_START_TIME
+     * @see Connectors#setRequestStartTime(HttpServerExchange)
      */
     public long getRequestStartTime() {
         return requestStartTime;
@@ -1757,18 +1764,18 @@ public final class HttpServerExchange extends AbstractAttachable {
     /**
      * Transmit the response headers. After this method successfully returns,
      * the response channel may become writable.
-     * <p/>
+     * <p>
      * If this method fails the request and response channels will be closed.
-     * <p/>
+     * <p>
      * This method runs asynchronously. If the channel is writable it will
      * attempt to write as much of the response header as possible, and then
      * queue the rest in a listener and return.
-     * <p/>
+     * <p>
      * If future handlers in the chain attempt to write before this is finished
      * XNIO will just magically sort it out so it works. This is not actually
      * implemented yet, so we just terminate the connection straight away at
      * the moment.
-     * <p/>
+     * <p>
      * TODO: make this work properly
      *
      * @throws IllegalStateException if the response headers were already sent
@@ -1858,6 +1865,10 @@ public final class HttpServerExchange extends AbstractAttachable {
         return ret;
     }
 
+    boolean isResumed() {
+        return anyAreSet(state, FLAG_SHOULD_RESUME_WRITES | FLAG_SHOULD_RESUME_READS);
+    }
+
     private static class ExchangeCompleteNextListener implements ExchangeCompletionListener.NextListener {
         private final ExchangeCompletionListener[] list;
         private final HttpServerExchange exchange;
@@ -1930,10 +1941,10 @@ public final class HttpServerExchange extends AbstractAttachable {
 
     /**
      * Channel implementation that is actually provided to clients of the exchange.
-     * <p/>
+     * <p>
      * We do not provide the underlying conduit channel, as this is shared between requests, so we need to make sure that after this request
      * is done the the channel cannot affect the next request.
-     * <p/>
+     * <p>
      * It also delays a wakeup/resumesWrites calls until the current call stack has returned, thus ensuring that only 1 thread is
      * active in the exchange at any one time.
      */
@@ -2101,10 +2112,10 @@ public final class HttpServerExchange extends AbstractAttachable {
      * Channel implementation that is actually provided to clients of the exchange. We do not provide the underlying
      * conduit channel, as this will become the next requests conduit channel, so if a thread is still hanging onto this
      * exchange it can result in problems.
-     * <p/>
+     * <p>
      * It also delays a readResume call until the current call stack has returned, thus ensuring that only 1 thread is
      * active in the exchange at any one time.
-     * <p/>
+     * <p>
      * It also handles buffered request data.
      */
     private final class ReadDispatchChannel extends DetachableStreamSourceChannel implements StreamSourceChannel {
@@ -2432,6 +2443,6 @@ public final class HttpServerExchange extends AbstractAttachable {
 
     @Override
     public String toString() {
-        return "HttpServerExchange{ " + getRequestMethod().toString() + " " + getRequestURI() + " request " + requestHeaders + " response " + responseHeaders + '}';
+        return "HttpServerExchange{ " + getRequestMethod().toString() + " " + getRequestURI() + '}';
     }
 }
